/*
 * Copyright [2013-2021], Alibaba Group Holding Limited
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.alibaba.polardbx.repo.mysql.handler.ddl.newengine;

import com.alibaba.polardbx.common.ddl.newengine.DdlState;
import com.alibaba.polardbx.common.exception.TddlRuntimeException;
import com.alibaba.polardbx.common.exception.code.ErrorCode;
import com.alibaba.polardbx.executor.cursor.Cursor;
import com.alibaba.polardbx.executor.cursor.impl.AffectRowCursor;
import com.alibaba.polardbx.executor.spi.IRepository;
import com.alibaba.polardbx.gms.metadb.misc.DdlEngineRecord;
import com.alibaba.polardbx.optimizer.context.ExecutionContext;
import com.alibaba.polardbx.optimizer.core.rel.dal.LogicalDal;
import org.apache.calcite.sql.SqlPauseDdlJob;

import java.util.List;

public class DdlEnginePauseJobsHandler extends DdlEngineJobsHandler {

    public DdlEnginePauseJobsHandler(IRepository repo) {
        super(repo);
    }

    @Override
    public Cursor doHandle(final LogicalDal logicalPlan, ExecutionContext executionContext) {
        SqlPauseDdlJob command = (SqlPauseDdlJob) logicalPlan.getNativeSqlNode();
        return doPause(command.isAll(), command.getJobIds(), executionContext);
    }

    public Cursor doPause(boolean isAll, List<Long> jobIds, ExecutionContext executionContext) {
        List<DdlEngineRecord> records =
            fetchRecords(executionContext.getSchemaName(), isAll, jobIds);
        records.stream().forEach(e -> {
            DdlState state = DdlState.valueOf(e.state);
            if (!(state == DdlState.RUNNING || state == DdlState.ROLLBACK_RUNNING)) {
                throw new TddlRuntimeException(ErrorCode.ERR_GMS_GENERIC, String.format(
                    "Only RUNNING/ROLLBACK_RUNNING jobs can be paused, job %s is in %s state", e.jobId, e.state));
            }
        });

        int countDone = 0;
        for (DdlEngineRecord record : records) {
            if (DdlState.RUNNING == DdlState.valueOf(record.state)) {
                if (schedulerManager.tryPauseDdl(
                    record.jobId,
                    DdlState.RUNNING,
                    DdlState.PAUSED)) {
                    countDone++;
                    interruptJob(record.schemaName, record.jobId);
                }
            } else if (DdlState.ROLLBACK_RUNNING == DdlState.valueOf(record.state)) {
                if (schedulerManager.tryPauseDdl(
                    record.jobId,
                    DdlState.ROLLBACK_RUNNING,
                    DdlState.ROLLBACK_PAUSED)) {
                    countDone++;
                    interruptJob(record.schemaName, record.jobId);
                }
            }
        }

        return new AffectRowCursor(new int[] {countDone});
    }

}
